use crate::render::scene::Scene;
use vulkano::command_buffer::{
    AutoCommandBufferBuilder, CommandBufferUsage, RenderingAttachmentInfo, RenderingInfo,
};
use vulkano::pipeline::graphics::viewport::Viewport;
use vulkano::render_pass::{AttachmentLoadOp, AttachmentStoreOp};
use vulkano::swapchain::{PresentMode, SwapchainPresentInfo};
use vulkano::sync::GpuFuture;
use vulkano::{Validated, VulkanError, sync};
use vulkano_util::context::VulkanoContext;
use vulkano_util::window::{VulkanoWindows, WindowDescriptor};
use winit::application::ApplicationHandler;
use winit::event::WindowEvent;
use winit::event_loop::ActiveEventLoop;
use winit::window::WindowId;

use super::vulkan_context::VulkanContext;

pub struct App {
    vulkan_context: VulkanContext,
    vulkano_windows: VulkanoWindows,
    scene: Option<Scene>,
}

impl From<VulkanoContext> for App {
    fn from(vulkano_context: VulkanoContext) -> Self {
        Self {
            vulkan_context: VulkanContext::new(vulkano_context),
            vulkano_windows: VulkanoWindows::default(),
            scene: None,
        }
    }
}

impl ApplicationHandler for App {
    fn resumed(&mut self, event_loop: &ActiveEventLoop) {
        self.vulkano_windows.create_window(
            event_loop,
            self.vulkan_context.vulkano_context(),
            &WindowDescriptor {
                title: "Rust ASH Test".to_string(),
                width: 800.0,
                height: 600.0,
                present_mode: PresentMode::Fifo,
                ..Default::default()
            },
            |_| {},
        );

        self.scene = Some(
            Scene::load(
                &self.vulkan_context,
                &self.vulkano_windows.get_primary_renderer_mut().unwrap(),
            )
            .unwrap(),
        );
    }

    fn window_event(&mut self, event_loop: &ActiveEventLoop, _id: WindowId, event: WindowEvent) {
        match event {
            WindowEvent::CloseRequested => {
                log::debug!("The close button was pressed; stopping");
                event_loop.exit();
            }
            WindowEvent::RedrawRequested => {
                let renderer = self.vulkano_windows.get_primary_renderer_mut().unwrap();
                let acquire_future = renderer.acquire(None, |_| {}).unwrap();

                let mut builder = AutoCommandBufferBuilder::primary(
                    self.vulkan_context.command_buffer_allocator().clone(),
                    self.vulkan_context
                        .vulkano_context()
                        .graphics_queue()
                        .queue_family_index(),
                    CommandBufferUsage::OneTimeSubmit,
                )
                .unwrap();

                {
                    let viewport = Viewport {
                        offset: [0.0, 0.0],
                        extent: renderer.resolution(),
                        depth_range: 0.0..=1.0,
                    };

                    builder
                        .begin_rendering(RenderingInfo {
                            color_attachments: vec![Some(RenderingAttachmentInfo {
                                load_op: AttachmentLoadOp::Clear,
                                store_op: AttachmentStoreOp::Store,
                                clear_value: Some([0.0, 0.0, 0.0, 1.0].into()),
                                ..RenderingAttachmentInfo::image_view(
                                    renderer.swapchain_image_view().clone(),
                                )
                            })],
                            ..Default::default()
                        })
                        .unwrap()
                        .set_viewport(0, [viewport].into_iter().collect())
                        .unwrap();
                }

                if let Some(scene) = self.scene.as_ref() {
                    scene
                        .render(&self.vulkan_context, &renderer, &mut builder)
                        .unwrap();
                }

                builder.end_rendering().unwrap();

                let command_buffer = builder.build().unwrap();

                let future = acquire_future
                    .then_execute(
                        self.vulkan_context
                            .vulkano_context()
                            .graphics_queue()
                            .clone(),
                        command_buffer,
                    )
                    .unwrap();

                renderer.present(future.boxed(), true);
            }
            _ => {}
        }
    }

    fn about_to_wait(&mut self, _event_loop: &ActiveEventLoop) {
        let window = self.vulkano_windows.get_primary_window().unwrap();
        window.request_redraw();
    }
}
